// Copyright 2000-2005 the Contributors, as shown in the revision logs.
// Licensed under the Apache Public Source License 2.0 ("the License").
// You may not use this file except in compliance with the License.

package org.ibex.mail;
import org.ibex.crypto.*;
import org.ibex.util.*;
import org.ibex.mail.protocol.*;
import org.ibex.js.*;
import org.ibex.io.*;
import org.ibex.io.Fountain;
import java.util.*;
import java.net.*;
import java.io.*;

// FIXME: body constraints (how to enforce?)
//   - messages must NEVER contain 8-bit binary data; this is a violation of IMAP
//   - RFC822 1,000-char limit per line [soft line limit (suggested): 78 chars /  hard line limit: 998 chars]

// FEATURE: PGP-signature-parsing
// FEATURE: mailing list header parsing (?)
// FEATURE: threading as in http://www.jwz.org/doc/threading.html

/** 
 *  [immutable] This class encapsulates a message "floating in the
 *  ether": RFC2822 data but no storage-specific flags or other
 *  metadata.
 */
public class Message extends MIME.Part {

    // Parsed Headers //////////////////////////////////////////////////////////////////////////////

    public final Address     to;
    public final Address     from;                // if multiple From entries, this is sender
    public final Address     envelopeFrom;
    public final Address     envelopeTo;
    public final Date        date;
    public final Date        arrival;
    public final Address     replyto;             // if none provided, this is equal to sender
    public final String      subject;
    public final String      messageid;
    public final Address[]   cc;
    public final Address[]   bcc;

    public static Message newMessage(Fountain in) throws Malformed { return new Message(in); }

    /*
    public Message reply(Fountain in, Address from, boolean includeReInSubject) throws Malformed {
	Address to = null;
	if (to==null) to = Address.parse(headers.get("reply-to"));
	if (to==null) to = Address.parse(headers.get("from"));
	if (to==null) to = envelopeFrom;
        if (to==null) throw new Malformed("cannot reply to a message without a return address");
	Message ret = newMessage(in, from, to);
	ret.headers.put("In-Reply-To", messageid);
	String references = headers.get("references");
	ret.headers.put("References", messageid + (references==null?"":(" "+references)));
	if (includeReInSubject && subject!=null && !subject.toLowerCase().trim().startsWith("re:"))
	    headers.put("subject", "Re: "+subject);
	return ret;
    }
    */

    // FIXME
    //public static Message newMessage(Headers headers, Fountain body, Address from, Address to) throws Malformed {
    //}

    public static Message newMessage(Fountain in, Address from, Address to) throws Malformed {
        StringBuffer sb = new StringBuffer();
        if (from != null) sb.append("Return-Path: " + from.toString(true) + "\r\n");
        Stream stream = in.getStream();
        while(true) {
            String s = stream.readln();
            if (s == null || s.length() == 0) {
                if (to != null) sb.append("Envelope-To: " + to.toString(true) + "\r\n");
                sb.append("\r\n");
                break;
            }
            if (to != null && s.toLowerCase().startsWith("envelope-to:")) continue;
            if (s.toLowerCase().startsWith("return-path:")) continue;
            sb.append(s);
            sb.append("\r\n");
        }
        for(String s = stream.readln(); s != null; s = stream.readln()) {
            sb.append(s);
            sb.append("\r\n");
        }
        return new Message(new Fountain.StringFountain(sb.toString()));
    }

    private Message(Fountain in) throws Malformed {
        super(in);

        this.envelopeTo   = headers.get("Envelope-To") != null ? Address.parse(headers.get("Envelope-To")) : null;
        this.envelopeFrom = headers.get("Return-Path") != null ? Address.parse(headers.get("Return-Path")) : null;
        this.to           = headers.get("To") != null ? Address.parse(headers.get("To")) : this.envelopeTo;
        this.from         = headers.get("From") != null ? Address.parse(headers.get("From")) : this.envelopeFrom;
        this.replyto      = headers.get("Reply-To") == null ? null : Address.parse(headers.get("Reply-To"));
        this.subject      = headers.get("Subject");
        this.messageid    = headers.get("Message-Id");
        this.cc           = Address.list(headers.get("Cc"));
        this.bcc          = Address.list(headers.get("Bcc"));
        this.date         = parseDate(headers.get("Date")) == null ? new Date() : parseDate(headers.get("Date"));

        // reenable this once whitelisting is moved out of javascript
        //if (this.messageid==null)
        //throw new RuntimeException("every RFC2822 message must have a Message-ID: header");

        /*
        // synthesize a message-id if not provided
        this.messageid    = headers.get("Message-Id") == null ? generateFreshMessageId(sha1(in.getStream())) : headers.get("Message-Id");
        if (headers.get("Message-Id") == null) {
            headers = headers.set("Message-Id", this.messageid);
            Log.warn(Message.class, "synthesizing message-id for " + summary());
        }
        */

        this.arrival      = this.date; // FIXME wrong; grab this from traces?
    }

    /*
    private static String sha1(Stream stream) {
        SHA1 sha1 = new SHA1();
        byte[] b = new byte[1024];
        while(true) {
            int numread = stream.read(b, 0, b.length);
            if (numread == -1) break;
            sha1.update(b, 0, numread);
        }
        byte[] results = new byte[sha1.getDigestSize()];
        sha1.doFinal(results, 0);
        return new String(Encode.toBase64(results));
    }
    */


    // Helpers /////////////////////////////////////////////////////////////////////////////

    // http://www.jwz.org/doc/mid.html
    private static final Random random = new Random();
    public static String generateFreshMessageId() {
        return generateFreshMessageId(Base36.encode(System.currentTimeMillis())+'.'+
                                      Base36.encode(random.nextLong()));
    }
    public static String generateFreshMessageId(String seed) {
        StringBuffer ret = new StringBuffer();
        ret.append('<');
        ret.append(seed);
        ret.append('@');
        try { ret.append(InetAddress.getLocalHost().getHostName()); } catch (UnknownHostException e) { /* DELIBERATE */ }
        ret.append('>');
        return ret.toString();
    }

    public static Date parseDate(String s) {
        // FIXME!!! this must be robust
        // date/time parsing: see spec, 3.3
        return null;
    }

    // this is belived to be compliant with QSBMF (http://cr.yp.to/proto/qsbmf.txt)
    public Message bounce(String reason) {
        if (envelopeFrom==null || envelopeFrom.toString().equals("")) return null;

        Log.warn(Message.class, "bouncing message due to: " + reason);
        Headers h = new Headers(headers, new String[] {
            "Envelope-To", envelopeFrom.toString(),
            "Return-Path", "<>",
            "From",        "MAILER-DAEMON <>",
            "To",          envelopeFrom.toString(),
            "Subject",     "failure notice"
        });

        String error =
            "\r\n"+
            "Hi. This is the Ibex Mail Server.  I'm afraid I wasn't able to deliver\r\n"+
            "your message to the following addresses. This is a permanent error;\r\n"+
            "I've given up.  Sorry it didn't work out\r\n."+
            "\r\n"+
            "<"+envelopeTo.toString()+">:\r\n"+
            reason+"\r\n"+
            "\r\n"+
            "--- Below this line is a copy of the message.\r\n"+
            "\r\n";

        try {
            return newMessage(Fountain.Util.concat(h, Fountain.Util.create(error), getBody()));
        } catch (Message.Malformed e) {
            Log.error(this, "caught Message.Malformed in Message.bounce(); this should never happen");
            Log.error(this, e);
            return null;
        }
    }

    public       String toString() { throw new RuntimeException("Message.toString() called"); }
    public final String summary() { return "[" + envelopeFrom + " -> " + envelopeTo + "] " + subject; }

    public static class Malformed extends MailException { public Malformed(String s) { super(s); } }
}

